/*
 * This file contains the bases of an interval arithmetic package (one more :).
 * Sylvain.Pion@sophia.inria.fr '97.
 *
 * The choice made is to represent an interval by a struct "DIS" of 2 doubles.
 * "i" stores the opposite of the lower bound.
 * "s" stores the upper bound.
 *
 * NOTE: Most functions in this file assume that the FP rounding mode
 *       is set to plus infinity.
 */

#ifndef DIS_H
#define DIS_H

/* Needed for DIS_sqrt() */
#include <math.h>
#include "FPU.h"

typedef struct {
  double i, s;		/* floats are almost always too limited in exponent. */
} DIS;			/* What about long double's ? */

static inline double dmax (double a, double b);	/* max(a,b) */
static inline double DIS_2_dbl (DIS a);		/* Cast to a double (middle) */

/* Basic operations. */

static inline DIS DIS_neg (DIS a);		/* Opposite */
static inline DIS DIS_dbl (double a);		/* Cast of a double */
static inline DIS DIS_add (DIS a, DIS b);	/* Addition */
static inline DIS DIS_sub (DIS a, DIS b);	/* Subtraction */
static inline DIS DIS_mul (DIS a, DIS b);	/* Multiplication */
static inline DIS DIS_squ (DIS a);		/* Square */

/* A little bit more semantic is needed for the following functions.
 * (it depends on your needs)
 */
static inline DIS DIS_sqrt (DIS a);		/* Square root */
static inline DIS DIS_div  (DIS a, DIS b);	/* Division */
static inline DIS DIS_abs  (DIS a);		/* Absolute value (needed?) */

/* I propose strict semantic for the following functions, that is:
 * - a positive answer means we are sure concerning the *value*.
 * - a negative answer means "no" or "maybe" (maybe = interval overlap).
 */

static inline int DIS_inf   (DIS a, DIS b);	/* a<b */
static inline int DIS_sup   (DIS a, DIS b);	/* a>b */
static inline int DIS_infeq (DIS a, DIS b);	/* a<=b */
static inline int DIS_supeq (DIS a, DIS b);	/* a>=b */
static inline int DIS_equ   (DIS a, DIS b);	/* a==b */
static inline int DIS_dif   (DIS a, DIS b);	/* a!=b */

/* For the moment: (+/-1 if sure, else 0)
 * Could be {-2,-1,0,1,2,3} to have more information ?
 * Or 2 separate functions ?
 */

enum { NEG = -1, UNK, POS };
/* enum { NEG = -2, NEG_ZERO, ZERO, POS_ZERO, POS, UNK }; */

static inline int DIS_sgn  (DIS a);		/* "Sign" */
static inline int DIS_lsgn (DIS a);		/* NEG:<=0, POS:>=0, UNK */
/* static inline int DIS_fsgn (DIS a); */	/* full info */


/* Here comes the code. */

static inline int DIS_sgn (DIS a)
{
  if (a.i<0) return POS;
  else if (a.s<0) return NEG;
  else return UNK;
}

static inline int DIS_lsgn (DIS a)
{
  if (a.i<=0) return POS; 
  else if (a.s<=0) return NEG;
  else return UNK;
}

static inline DIS DIS_dbl (double a)
{
  DIS res;
  res.i = -a;
  res.s = a;
  return res;
}

static inline DIS DIS_neg (DIS a)
{
  DIS res;
  res.i = a.s;
  res.s = a.i;
  return res;
}

static inline DIS DIS_add (DIS a, DIS b)
{
  DIS res;
  res.i = a.i + b.i;
  res.s = a.s + b.s;
  return res;
}

static inline DIS DIS_sub (DIS a, DIS b)
{
  DIS res;
  res.i = a.i + b.s;
  res.s = a.s + b.i;
  return res;
}

static inline DIS DIS_squ (DIS a)
{
  DIS res;
  if (a.i<=0)
  {
    res.i = (-a.i)*a.i;
    res.s = a.s*a.s;
  }
  else if (a.s<=0)
  {
    res.i = (-a.s)*a.s;
    res.s = a.i*a.i;
  }
  else 
  {
    res.i = 0;
    if (a.i<a.s) res.s = a.s*a.s;
    else         res.s = a.i*a.i;
  };
  return res;
}

static inline DIS DIS_mul (DIS a, DIS b)
{
  DIS res;
  if (a.i<=0)					/* a>=0 */
  {
    if (b.i<=0)					/* b>=0 */
    {
      res.i = (-a.i)*b.i;
      res.s = a.s*b.s;
    }
    else if (b.s<=0)				/* b<=0 */
    {
      res.i = a.s*b.i;
      res.s = (-a.i)*b.s;
    }
    else					/* 0 \in b */
    {
      res.i = a.s*b.i;
      res.s = a.s*b.s;
    };
  }
  else if (a.s<=0)				/* a<=0 */
  {
    if (b.i<=0)                                 /* b>=0 */
    {
      res.i = a.i*b.s;
      res.s = a.s*(-b.i);
    }
    else if (b.s<=0)                            /* b<=0 */
    {
      res.i = (-a.s)*b.s;
      res.s = a.i*b.i;
    }
    else                                        /* 0 \in b */
    {
      res.i = a.i*b.s;
      res.s = a.i*b.i;
    };
  }
  else						/* 0 \in [a.i;a.s] */
  {
    if (b.i<=0)                                 /* b>=0 */
    {
      res.i = a.i*b.s;
      res.s = a.s*b.s;
    }
    else if (b.s<=0)                            /* b<=0 */
    {
      res.i = a.s*b.i;
      res.s = a.i*b.i;
    }
    else                                        /* 0 \in b */
    {
      double tmp1, tmp2;
      tmp1 = a.i*b.s;
      tmp2 = a.s*b.i;
      res.i = (tmp1>tmp2) ? tmp1 : tmp2;
      tmp1 = a.i*b.i;
      tmp2 = a.s*b.s;
      res.s = (tmp1>tmp2) ? tmp1 : tmp2;
    };
  };
  return res;
}

static inline DIS DIS_sqrt (DIS a)
{
  DIS res;
  FPU_set_rounding_to_minus_infinity();
  res.i = -sqrt(-a.i);
  FPU_set_rounding_to_infinity();
  res.s = sqrt(a.s);
  return res;
}

static inline double dmax (double a, double b)
{
  return (a>b) ? a : b;
}

static inline DIS DIS_abs (DIS a)
{
  DIS res;
  if (a.i>0)
  {
    if (a.s>0)
    {
      res.i = 0;
      res.s = dmax(a.i,a.s);
    }
    else
      res = DIS_neg(a);
  }
  else
    res = a;
  return res;
}

static inline DIS DIS_div (DIS a, DIS b)
{
  DIS res;
  if (b.i<0)				/* b>0 */
  {
    if (a.i<=0)				/* a>=0 */
    {
      res.i = a.i/b.s;
      res.s = a.s/(-b.i);
    }
    else if (a.s<=0)			/* a<=0 */
    {
      res.i = a.i/(-b.i);
      res.s = a.s/b.s;
    }
    else				/* 0 \in a */
    {
      res.i = a.i/(-b.i);
      res.s = a.s/(-b.i);
    };
  }
  else if (b.s<0)			/* b<0 */
  {
    if (a.i<=0)                         /* a>=0 */
    {
      res.i = a.s/(-b.s);
      res.s = a.i/b.i;
    }
    else if (a.s<=0)                    /* a<=0 */
    {
      res.i = a.s/b.i;
      res.s = a.i/(-b.s);
    }
    else                                /* 0 \in a */
    {
      res.i = a.s/(-b.s);
      res.s = a.i/(-b.s);
    };
  }
  else					/* 0 \in [b.i;b.s] :-) */
  {
    res.i = res.s = HUGE_VAL;
    if (a.i>=0)                         /* a>=0 */
    { }
    else if (a.s<=0)                    /* a<=0 */
    { }
    else                                /* 0 \in a */
    { };
  };
  return res;
}

static inline int DIS_inf (DIS a, DIS b)
{
  return a.s < -b.i;
}

static inline int DIS_sup (DIS a, DIS b)
{
  return b.s < -a.i;
}

static inline int DIS_infeq (DIS a, DIS b)
{
  return a.s <= -b.i;
}

static inline int DIS_supeq (DIS a, DIS b)
{
  return b.s <= -a.i;
}

static inline int DIS_equ (DIS a, DIS b)
{
  return (-a.i == a.s) && (-b.i == b.s) && (a.s == b.s);
}

static inline int DIS_dif (DIS a, DIS b)
{
  return (-a.i != a.s) || (-b.i != b.s) || (a.s != b.s);
}

static inline double DIS_2_dbl (DIS a)
{
  return (a.s-a.i)*.5;
}

#endif
