
// This version is kept uptodate with the one in the CGAL library.
// They differ only by the copyright license.
// Author:  Sylvain.Pion@sophia.inria.fr  1997-1998.
//
// This file contains the description of the two classes:
// - Dis_advanced  (do the FPU rounding mode changes yourself)
// - Dis	   ("plug-in" version, derived from the other one)
// The second one is _much_ slower (till a factor 10).

#ifndef Dis_H
#define Dis_H

#include <iostream.h>
#include <math.h>
#include "FPU.h"	// FPU rounding mode functions.

class Dis_advanced
{
  friend Dis_advanced sqrt(const Dis_advanced&);

public:
  Dis_advanced(const double i, const double s)
  {
    inf = -i;
    sup = s;
  }

  Dis_advanced(const double d = 0.0)
  {
    inf = -d;
    sup = d;
  }

  Dis_advanced& operator=(const Dis_advanced& d)
  {
    inf = d.inf;
    sup = d.sup;
    return *this;
  }

  Dis_advanced operator+(const Dis_advanced& d) const
  {
    return Dis_advanced(-(inf + d.inf), sup + d.sup);
  }

  Dis_advanced& operator+=(const Dis_advanced& d)
  {
    inf += d.inf;
    sup += d.sup;
    return *this;
  }

  Dis_advanced operator-(const Dis_advanced& d) const
  {
    return Dis_advanced(-(inf + d.sup), sup + d.inf);
  }

  Dis_advanced& operator-=(const Dis_advanced& d)
  {
    inf += d.sup;
    sup += d.inf;
    return *this;
  }

  Dis_advanced operator*(const Dis_advanced& d) const
  {
    if (inf<=0)                                   /* this>=0 */
    {
      if (d.inf<=0)                                 /* d>=0 */
	return Dis_advanced(-((-inf)*d.inf), sup*d.sup);
      else if (d.sup<=0)                            /* d<=0 */
	return Dis_advanced(-(sup*d.inf), (-inf)*d.sup);
      else                                        /* 0 \in d */
	return Dis_advanced(-(sup*d.inf), sup*d.sup);
    }
    else if (sup<=0)                              /* this<=0 */
    {
      if (d.inf<=0)                                 /* d>=0 */
	return Dis_advanced(-(inf*d.sup), sup*(-d.inf));
      else if (d.sup<=0)                            /* d<=0 */
	return Dis_advanced(-((-sup)*d.sup), inf*d.inf);
      else                                        /* 0 \in d */
	return Dis_advanced(-(inf*d.sup), inf*d.inf);
    }
    else                                          /* 0 \in [inf;sup] */
    {
      if (d.inf<=0)                                 /* d>=0 */
	return Dis_advanced(-(inf*d.sup), sup*d.sup);
      else if (d.sup<=0)                            /* d<=0 */
	return Dis_advanced(-(sup*d.inf), inf*d.inf);
      else                                        /* 0 \in d */
      {
        double tmp1, tmp2, tmp3, tmp4;
        tmp1 = inf*d.sup;
        tmp2 = sup*d.inf;
        tmp3 = inf*d.inf;
        tmp4 = sup*d.sup;
	return Dis_advanced(-((tmp1>tmp2) ? tmp1 : tmp2),
					   (tmp3>tmp4) ? tmp3 : tmp4);
      };
    };
  }

  Dis_advanced & operator*=(const Dis_advanced& d)
  {
    *this = *this * d;
    return *this;
  }

  Dis_advanced operator/(const Dis_advanced& d) const
  {
    if (d.inf<0.0)                            /* d>0 */
    {
      if (inf<=0.0)                         /* this>=0 */
	return Dis_advanced(-(inf/d.sup), sup/(-d.inf));
      else if (sup<=0.0)                    /* this<=0 */
	return Dis_advanced(-(inf/(-d.inf)), sup/d.sup);
      else                                /* 0 \in this */
	return Dis_advanced(-(inf/(-d.inf)), sup/(-d.inf));
    }
    else if (d.sup<0.0)                       /* d<0 */
    {
      if (inf<=0.0)                         /* this>=0 */
	return Dis_advanced(-(sup/(-d.sup)), inf/d.inf);
      else if (sup<=0.0)                    /* b<=0 */
	return Dis_advanced(-(sup/d.inf), inf/(-d.sup));
      else                                /* 0 \in this */
	return Dis_advanced(-(sup/(-d.sup)), inf/(-d.sup));
    }
    else                                  /* 0 \in [d.inf;d.sup] */
      return Dis_advanced(-(HUGE_VAL), HUGE_VAL);
  }

  Dis_advanced & operator/=(const Dis_advanced& d)
  {
    *this = *this / d;
    return *this;
  }

  Dis_advanced operator-() const
  {
    return Dis_advanced(-(sup), inf);
  }

  bool operator==(const Dis_advanced& d) const
  {
    return overlap(d);
  }

  bool operator!=(const Dis_advanced& d) const
  {
    return !overlap(d);
  }

  bool operator<(const Dis_advanced& d) const
  {
    return (sup < -d.inf);
  }

  bool operator>(const Dis_advanced& d) const
  {
    return (d.sup < -inf);
  }

  bool operator<=(const Dis_advanced& d) const
  {
    return !(*this > d);
  }

  bool operator>=(const Dis_advanced& d) const
  {
    return !(*this < d);
  }

  double lower_bound() const
  {
    return -inf;
  }

  double upper_bound() const
  {
    return sup;
  }

private:
  bool overlap(const Dis_advanced& d) const
  {
    return (sup >= -d.inf) && (d.sup >= -inf);
  }

protected:
        // "inf" stores the OPPOSITE of the lower bound.
        // "sup" stores the upper bound of the interval.
  double inf, sup;
};


class Dis : public Dis_advanced
{
  friend Dis sqrt(const Dis&);

public:

  Dis(const double d = 0.0)
      : Dis_advanced(d) {}
  Dis(const double a, const double b)
      : Dis_advanced(a,b) {}

  Dis(const Dis_advanced &d)
    : Dis_advanced(d) {}

  Dis operator+(const Dis& d) const
  {
    FPU_set_rounding_to_infinity();
    Dis tmp (-(inf + d.inf), sup + d.sup);
    FPU_set_rounding_to_nearest();
    return tmp;
  }

  Dis& operator+=(const Dis& d)
  {
    FPU_set_rounding_to_infinity();
    inf += d.inf;
    sup += d.sup;
    FPU_set_rounding_to_nearest();
    return *this;
  }

  Dis operator-(const Dis& d) const
  {
    FPU_set_rounding_to_infinity();
    Dis tmp (-(inf + d.sup), sup + d.inf);
    FPU_set_rounding_to_nearest();
    return tmp;
  }

  Dis& operator-=(const Dis& d)
  {
    FPU_set_rounding_to_infinity();
    inf += d.sup;
    sup += d.inf;
    FPU_set_rounding_to_nearest();
    return *this;
  }

  Dis operator*(const Dis& d) const
  {
    FPU_set_rounding_to_infinity();
    Dis tmp = ((Dis_advanced) *this)
	* (Dis_advanced) d;
    FPU_set_rounding_to_nearest();
    return tmp;
  }

  Dis& operator*=(const Dis& d)
  {
    FPU_set_rounding_to_infinity();
    *this = ((Dis_advanced) *this)
	* (Dis_advanced) d;
    FPU_set_rounding_to_nearest();
    return *this;
  }

  Dis operator/(const Dis& d) const
  {
    FPU_set_rounding_to_infinity();
    Dis tmp = ((Dis_advanced) *this)
	/ (Dis_advanced) d;
    FPU_set_rounding_to_nearest();
    return tmp;
  }

  Dis& operator/=(const Dis& d)
  {
    FPU_set_rounding_to_infinity();
    *this = ((Dis_advanced) *this)
	/ (Dis_advanced) d;
    FPU_set_rounding_to_nearest();
    return *this;
  }
};


inline Dis_advanced sqrt(const Dis_advanced& d)
{
  FPU_set_rounding_to_minus_infinity();
  Dis_advanced tmp ( sqrt(-d.inf), 0);
  FPU_set_rounding_to_infinity();
  tmp.sup = sqrt(d.sup);
  return tmp;
}

inline Dis sqrt(const Dis& d)
{
  Dis tmp = sqrt( (Dis_advanced) d);
  FPU_set_rounding_to_nearest();
  return tmp;
}

ostream& operator<<(ostream& os, Dis_advanced& d)
{
  os << "[" << d.lower_bound() << ";" << d.upper_bound() << "]";
  return os;
}

#endif /* Dis_H */
