#ifndef FPU_H
#define FPU_H

/*
 * Specify some *very* platform dependant functions,
 * regarding the FPU rounding modes.
 * The information was found in the include files of the different systems,
 * and especially in those of the GNU libc 2.0.5.  Also in packages like BIAS.
 * There is only support for double precision.
 *
 * The supported architectures are:
 * - Linux		i386,sparc,alpha
 * - Sun/Solaris	sparc	(See README for SunOS4)
 * - DEC/OSF		alpha	(See README)
 * - SGI/Irix		mips	(6.2 works, see README for 5.3)
 *
 * I try to make it work with gcc/g++/cc/CC.
 * And I also try to provide the equivalent assembly code for GNU C.
 *
 * Sylvain.Pion@sophia.inria.fr, 1997-1998.
 */

#ifdef __i386__
#ifndef __i386
#define __i386
#endif
#endif

#if (!defined(__i386)&&!defined(__sparc)&&!defined(__alpha)&&!defined(__mips))
#error "Architecture not recognized."
#endif

#if (defined(__GNUC__) && (defined(__i386) || defined(__sparc) || defined(__alpha)))
#define __USE_ASSEMBLY
#else
#ifdef __GNUC__
#  warning "Not using assembly in FPU.h"
#endif
#endif

#ifndef __USE_ASSEMBLY
#ifdef __linux
#include <fpu_control.h>
#endif
#if defined(__osf__) && !defined(__osf)
#define __osf
#endif
#ifdef __osf
#include </usr/include/float.h>
#endif
#ifdef __sgi
#include <sys/fpu.h>
#endif
#ifdef __sun
#include <ieeefp.h>
#endif
#else	/* __USE_ASSEMBLY */
#ifdef __i386
#define SETFPCW(CW) asm volatile ("fldcw %0" : : "g" (CW))
/* x86:                       rounding   | precision  | default mask */
static const unsigned int cw_zero = 0xC00      | 0x200      | 0x107f; 
static const unsigned int cw_near = 0x0        | 0x200      | 0x107f;
static const unsigned int cw_up   = 0x800      | 0x200      | 0x107f; 
static const unsigned int cw_down = 0x400      | 0x200      | 0x107f;
#endif
#ifdef __sparc
#define SETFPCW(CW) asm volatile ("ld %0,%%fsr" : : "g" (CW))
/* Sparc:                     rounding   | precision  | default mask */
static const unsigned int cw_zero = 0x40000000 | 0x20000000 | 0x1f;
static const unsigned int cw_near = 0x0        | 0x20000000 | 0x1f;
static const unsigned int cw_up   = 0x80000000 | 0x20000000 | 0x1f;
static const unsigned int cw_down = 0xc0000000 | 0x20000000 | 0x1f;
#endif
#ifdef __alpha
#define SETFPCW(CW) asm volatile ("mt_fpcr %0; excb" : : "f" (CW))
/* Alpha:                      rounding */
static const unsigned long cw_zero = 0x0000000000000000UL;
static const unsigned long cw_near = 0x0800000000000000UL;
static const unsigned long cw_up   = 0x0c00000000000000UL;
static const unsigned long cw_down = 0x0400000000000000UL;
#endif
#endif

#if !defined(__GNUC__) && !defined(__cplusplus)
#define inline /* */
#endif

static inline void FPU_set_rounding_to_zero (void);
static inline void FPU_set_rounding_to_nearest (void);
static inline void FPU_set_rounding_to_infinity (void);
static inline void FPU_set_rounding_to_minus_infinity (void);


/* Code of those inline functions: */

static inline void FPU_set_rounding_to_zero (void)
{
#ifdef __USE_ASSEMBLY
	SETFPCW(cw_zero);
#else
#ifdef __linux
#ifdef __i386 
        __setfpucw(0x1e72);
#else
        __setfpucw(0x1f72);
#endif
#endif

#ifdef __osf
	write_rnd(FP_RND_RZ);
#endif

#ifdef __sun
        fpsetround(FP_RZ);
#endif

#ifdef __sgi
	union fpc_csr fpu_ctl;
	fpu_ctl.fc_word = get_fpc_csr();
	fpu_ctl.fc_struct.rounding_mode = ROUND_TO_ZERO;
	set_fpc_csr (fpu_ctl.fc_word);
#endif
#endif
}

static inline void FPU_set_rounding_to_nearest (void)
{
#ifdef __USE_ASSEMBLY
	SETFPCW(cw_near);
#else
#ifdef __osf
	write_rnd(FP_RND_RN);
#endif

#ifdef __sun
	fpsetround(FP_RN);
#endif

#ifdef __linux
#ifdef __i386
	__setfpucw(0x1272);
#else
	__setfpucw(0x1372);
#endif
#endif

#ifdef __sgi
	union fpc_csr fpu_ctl;
	fpu_ctl.fc_word = get_fpc_csr();
	fpu_ctl.fc_struct.rounding_mode = ROUND_TO_NEAREST;
	set_fpc_csr (fpu_ctl.fc_word);
#endif
#endif
}

static inline void FPU_set_rounding_to_infinity (void)
{
#ifdef __USE_ASSEMBLY
	SETFPCW(cw_up);
#else
#ifdef __osf
	write_rnd(FP_RND_RP);
#endif

#ifdef __sun
        fpsetround(FP_RP);
#endif

#ifdef __linux
#ifdef __i386
	__setfpucw(0x1a72);
#else
	__setfpucw(0x1b72);
#endif
#endif

#ifdef __sgi
	union fpc_csr fpu_ctl;
	fpu_ctl.fc_word = get_fpc_csr();
	fpu_ctl.fc_struct.rounding_mode = ROUND_TO_PLUS_INFINITY;
	set_fpc_csr (fpu_ctl.fc_word);
#endif
#endif
}

static inline void FPU_set_rounding_to_minus_infinity (void)
{
#ifdef __USE_ASSEMBLY
	SETFPCW(cw_down);
#else
#ifdef __osf
	write_rnd(FP_RND_RM);
#endif

#ifdef __sun
        fpsetround(FP_RM);
#endif

#ifdef __linux
#ifdef __i386
        __setfpucw(0x1672);
#else
        __setfpucw(0x1772); /* FIX ME */
#endif
#endif

#ifdef __sgi
	union fpc_csr fpu_ctl;
	fpu_ctl.fc_word = get_fpc_csr();
	fpu_ctl.fc_struct.rounding_mode = ROUND_TO_MINUS_INFINITY;
	set_fpc_csr (fpu_ctl.fc_word);
#endif
#endif
}

#endif
